#!/usr/bin/env python

'''rise_set/sky_coordinates.py
Validated representations of RA and Dec.

Author: Michelle Becker (mbecker@lcogt.net)
April 2011
'''

# Required for true (non-integer) division
from __future__ import division
from builtins import str


# Internal imports
from rise_set.angle import Angle, InvalidAngleError


class RightAscension(Angle):
    def __init__(self, time=None, degrees=None, radians=None, units='time'):

        angle_degrees = time
        angle_radians = radians
        angle_units   = units

        if degrees != None:
            angle_degrees = degrees
            angle_units   = 'arc'

        Angle.__init__(self, angle_degrees, angle_radians, angle_units)

        # Check input
        self.validate_ra()


    def validate_ra(self):
        '''Check the right ascension is valid (0 <= ra < 24 00 00).'''

        # RA cannot be greater than one circle in size
        if ( self.degrees >= 360 or self.degrees < 0 ) :
            msg  = "'%s' is not within allowable RA range" % str(self.degrees)
            msg += " (0 <= ra < 360 degrees)."
            raise InvalidAngleError(msg)

        return True


    def __repr__(self):
        return "RightAscension(%s degrees)" % self.in_degrees()



class Declination(Angle):
    def __init__(self, degrees=None, radians=None, units='arc'):
        Angle.__init__(self, degrees, radians, units)

        # Check input
        self.validate_dec()


    def validate_dec(self):
        '''Check the declination is valid (-90 <= dec <= +90).'''

        # Treat the special case of the north and south poles
        if ( self.degrees > 90   or   self.degrees < -90 ):
            msg  = "'%s' is not within allowable Dec range" % str(self.degrees)
            msg += " (-90 <= dec <= +90 degrees)."
            raise InvalidAngleError(msg)

        return True


    def __repr__(self):
        return "Declination(%s degrees)" % self.in_degrees()



#!/usr/bin/env python

'''
test_utils.py - Unit tests for rise_set.utils

description

Author: Eric Saunders
December 2013
'''
from builtins import object

from rise_set.utils import (coalesce_adjacent_intervals, intersect_intervals,
                            intersect_many_intervals)

import datetime
from nose.tools import assert_equal


class TestUtils(object):

    def setup(self):
        self.some_adjacent_intervals = [
               # Two contiguous intervals
               (datetime.datetime(year=2011, month=2, day=1, hour=12, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=12, minute=30)),
               (datetime.datetime(year=2011, month=2, day=1, hour=12, minute=30),
                datetime.datetime(year=2011, month=2, day=1, hour=13, minute=0)),
               # Three contiguous intervals
               (datetime.datetime(year=2011, month=2, day=1, hour=15, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=15, minute=30)),
               (datetime.datetime(year=2011, month=2, day=1, hour=15, minute=30),
                datetime.datetime(year=2011, month=2, day=1, hour=16, minute=0)),
               (datetime.datetime(year=2011, month=2, day=1, hour=16, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=16, minute=30)),
               # Two non-contiguous intervals
               (datetime.datetime(year=2011, month=2, day=1, hour=17, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=17, minute=30)),
               (datetime.datetime(year=2011, month=2, day=1, hour=18, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=18, minute=30)),
              ]

        self.expected_coalescence = [
               # First coalescence
               (datetime.datetime(year=2011, month=2, day=1, hour=12, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=13, minute=0)),
               # Second coalescence
               (datetime.datetime(year=2011, month=2, day=1, hour=15, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=16, minute=30)),
               # Remaining, non-continous intervals
               (datetime.datetime(year=2011, month=2, day=1, hour=17, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=17, minute=30)),
               (datetime.datetime(year=2011, month=2, day=1, hour=18, minute=0),
                datetime.datetime(year=2011, month=2, day=1, hour=18, minute=30)),
            ]


    def test_coalesce_adjacent_intervals(self):
        input_intervals = list(self.some_adjacent_intervals)
        received = coalesce_adjacent_intervals(self.some_adjacent_intervals)

        assert_equal(received, self.expected_coalescence)

        # Confirm the input list was not modified in place
        assert_equal(input_intervals, self.some_adjacent_intervals)


    def test_intersect_case01(self):
        # case 1      |......|
        #     |.....|
        expected = []

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15,  0, 0, 0), datetime.datetime(2013, 6, 15,  3, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case02(self):
        # case 2      |......|
        #          |.....|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,14,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15,  9, 0, 0), datetime.datetime(2013, 6, 15, 14, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case03(self):
        # case 3      |......|
        #              |....|
        expected = [(datetime.datetime(2013,6,15,13,0,0),datetime.datetime(2013,6,15,14,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 13, 0, 0), datetime.datetime(2013, 6, 15, 14, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case04(self):
        # case 4      |......|
        #                 |.....|
        expected = [(datetime.datetime(2013,6,15,13,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 13, 0, 0), datetime.datetime(2013, 6, 15, 16, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case05(self):
        # case 5      |......|
        #                      |......|
        expected = []

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 16, 0, 0), datetime.datetime(2013, 6, 15, 18, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case06(self):
        # case 6      |......|
        #          |............|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 11, 0, 0), datetime.datetime(2013, 6, 15, 16, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)

## Edge cases ##

    def test_intersect_case07(self):
        # case 7      |......|
        #             |......|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case08(self):
        # case 8      |......|
        #      |......|
        expected = []

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15,  9, 0, 0), datetime.datetime(2013, 6, 15, 12, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case09(self):
        # case 9      |......|
        #                    |......|
        expected = []

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 15, 0, 0), datetime.datetime(2013, 6, 15, 18, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_case10(self):
        # case 10     |......|
        #             |........|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 16, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)

    def test_intersect_case11(self):
        # case 11     |......|
        #           |........|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 11, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)

## Multiple intersections

    def test_intersect_case12(self):
        # case 12     |......|
        #           |...| |....|
        expected = [(datetime.datetime(2013,6,15,12,0,0),datetime.datetime(2013,6,15,13,0,0)),
                    (datetime.datetime(2013,6,15,14,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 11, 0, 0), datetime.datetime(2013, 6, 15, 13, 0, 0)),
                (datetime.datetime(2013, 6, 15, 14, 0, 0), datetime.datetime(2013, 6, 15, 16, 0, 0))]

        received = intersect_intervals(int1, int2)

        assert_equal(received, expected)


    def test_intersect_many(self):
        # |......|
        #    |......|
        #      |......|
        expected = [(datetime.datetime(2013,6,15,14,0,0),datetime.datetime(2013,6,15,15,0,0))]

        int1 = [(datetime.datetime(2013, 6, 15, 12, 0, 0), datetime.datetime(2013, 6, 15, 15, 0, 0))]
        int2 = [(datetime.datetime(2013, 6, 15, 13, 0, 0), datetime.datetime(2013, 6, 15, 16, 0, 0))]
        int3 = [(datetime.datetime(2013, 6, 15, 14, 0, 0), datetime.datetime(2013, 6, 15, 17, 0, 0))]

        received = intersect_many_intervals(int1,int2,int3)
        assert_equal(received,expected)



#!/usr/bin/env python

'''
moving_objects.py - Routines for moving object (asteroids, comets) calculations

description

Authors: Eric Saunders
         Tim Lister
December 2013
'''

from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object

from rise_set.astrometry import (gregorian_to_ut_mjd, elem_to_topocentric_apparent,
                                 calc_local_hour_angle, calculate_altitude, date_to_tdb)
from rise_set.angle      import Angle
from rise_set.utils      import coalesce_adjacent_intervals, MovingViolation

import re

import ast
from datetime import datetime,timedelta


def initialelemdict():
    '''Create an inital empty orbital elements dictionary for use by read_neocp_orbit.'''
    keys  = "name H G epoch mean_anomaly long_node arg_perihelion inclination "
    keys += "eccentricity MDM semi_axis n_obs n_nights type"

    return {}.fromkeys(keys.split())


def extract_mpc_epoch(epochstring):
    '''Convert packed MPC epoch format (e.g. 'J974L') from NEOCP orbit files
    into a datetime.datetime epoch (e.g. '1997 4 21'). Returns -1 if invalid
    length (no other sanity checking is done).'''

    if len(epochstring) != 5: return -1
    year = 100 * (ord(epochstring[0]) - ord('A') + 10) + \
           10 * (ord(epochstring[1]) - ord('0')) +       \
           (ord(epochstring[2]) - ord('0'))

    arr = [0, 0]
    for i in range(0,2,+1):
        if (ord(epochstring[i + 3]) >= ord('A')):
            arr[i] = ord(epochstring[i + 3]) - ord('A') + 10
        else:
            arr[i] = ord(epochstring[i + 3]) - ord('0')

    return datetime(year, arr[0], arr[1], 0)


def read_neocp_orbit(orbfile):
    '''Read an NEOCP orbit file, storing the contents as a dict.
    Comet format is from
    http://www.minorplanetcenter.net/iau/info/CometOrbitFormat.html (Ephemerides
    and Orbital Elements Format), asteroid format is from battle-weary
    inference but is also documented at
    http://www.minorplanetcenter.net/iau/info/MPOrbitFormat.html'''

    comet_regex = r'^\d{4}P'
    comet_pat = re.compile(comet_regex)

    elements = initialelemdict()

    orb_fh = open(orbfile, 'r')
    for line in orb_fh.readlines():
        if line.lstrip()[0] == '<': continue
        if line[0:7] == ' REMARK' : continue
        if not len(line.strip())  : continue

        line   = line.rstrip()
        chunks = line.split()

# Check if it is a periodic (numbered) comet or long-period comet and
# set type/scheme appropriately
        if line.startswith('    C') or re.match(comet_pat, line):
            elements['type']           = 'MPC_COMET'
            is_comet = True
        else:
            elements['type']           = 'MPC_MINOR_PLANET'
            is_comet = False

        elements['name']           = chunks[0]

        if is_comet == True:
            epoch_of_perih = datetime(year=int(chunks[1]), month=int(chunks[2]), day=int(float(chunks[3])))
            epoch_of_perih = epoch_of_perih + timedelta(days=float(chunks[3])-int(float(chunks[3])))
            elements['epochofperih']   = gregorian_to_ut_mjd(epoch_of_perih)
            elements['perihdist']      = float(chunks[4])
            elements['eccentricity']   = float(chunks[5])
            elements['arg_perihelion'] = Angle(degrees=float(chunks[6]))
            elements['long_node']      = Angle(degrees=float(chunks[7]))
            elements['inclination']    = Angle(degrees=float(chunks[8]))
            epoch_dt = datetime.strptime(chunks[9], '%Y%m%d')
            elements['epoch']          = gregorian_to_ut_mjd(epoch_dt)
            elements['H']              = float(chunks[10])
            elements['G']              = float(chunks[11])
        else:
            elements['H']              = float(chunks[1])
            elements['G']              = float(chunks[2])

            epoch_dt                   = extract_mpc_epoch(chunks[3])
            elements['epoch']          = gregorian_to_ut_mjd(epoch_dt)

            elements['mean_anomaly']   = Angle(degrees=float(chunks[4]))
            elements['arg_perihelion'] = Angle(degrees=float(chunks[5]))
            elements['long_node']      = Angle(degrees=float(chunks[6]))
            elements['inclination']    = Angle(degrees=float(chunks[7]))
            elements['eccentricity']   = float(chunks[8])
            elements['MDM']            = Angle(degrees=float(chunks[9]))
            elements['semi_axis']      = float(chunks[10])

            # The next bit is...complicated... depending on whether it's a 
            # multi-opposition orbit or not
            # From http://www.minorplanetcenter.net/iau/info/MPOrbitFormat.html
            # 106        i1     Uncertainty parameter, U
            # 108 - 116  a9     Reference (chunks[12])
            # 118 - 122  i5     Number of observations
            # 124 - 126  i3     Number of oppositions
            #
            #    For multiple-opposition orbits:
            #    128 - 131  i4     Year of first observation
            #    132        a1     '-'
            #    133 - 136  i4     Year of last observation
            #
            #    For single-opposition orbits:
            #    128 - 131  i4     Arc length (days)
            #    133 - 136  a4     'days'
            #
            # 138 - 141  f4.2   r.m.s residual (")
            #
            # Extract a string section and turn this into an uncertainty
            # and a reference.
            uncertainty = line[105:106]
            if uncertainty == ' ':
                uncertainty = 'U'
            elements['uncertainty'] = uncertainty
            reference = line[107:116]
            if reference.strip() == '':
                reference = ''
            elements['reference'] = reference

            # multiple opposition
            opp_data = line[117:141]
            single_opp = False
            if 'days' in opp_data or 'min' in opp_data or 'hrs' in opp_data:
                single_opp = True 
            elements['n_obs']   = int(opp_data[0:5])
            elements['n_oppos'] = int(opp_data[6:9])
            try:
                elements['residual'] = float(opp_data[20:24])
            except ValueError:
                elements['residual'] = 9.99
            # If it's a single opposition orbit, return arc length
            if single_opp == True:
                if 'min' in opp_data:
                    num_nights_or_years = float(opp_data[10:14]) / 1440.0
                elif 'hrs' in opp_data:
                    num_nights_or_years = float(opp_data[10:14]) / 24.0
                else:
                    num_nights_or_years = int(opp_data[10:14])

            # ...or it's a multi-opposition orbit...
            else:
                num_nights_or_years = opp_data[10:19]

            elements['n_nights'] = num_nights_or_years
            # Try and retrieve last obs date
            if len(line) == 202:
                 elements['last_obs'] = line[194:202]


        # We only want the first non-header line
        break


    orb_fh.close()
    return elements



def file_to_dicts(filename):
    with open(filename, 'r') as fh:
        data = fh.read()
        return ast.literal_eval(data)


def initialise_sites(site_filename):
    '''Factory method to create a Sites object from a telescopes.dat file.'''
    site_dicts = file_to_dicts(site_filename)

    sites = Sites()
    sites.add_sites(site_dicts)

    return sites


class Sites(object):
    '''Collection for a group of site dictionaries. Only the first site with a
       given lat/long is stored. This allows us to conveniently read a
       file with a list of telescopes, which may share the same site.'''

    def __init__(self):
        self.sites = {}
        self.site_lat_longs = []


    def add_sites(self, site_dicts):
        for site_dict in site_dicts:
            self.add_site(site_dict)

        return


    def add_site(self, site_dict):
        lat_long = (site_dict['latitude'], site_dict['longitude'])
        HOURS_TO_DEGREES = 15.0

        if lat_long not in self.site_lat_longs:
            self.site_lat_longs.append(lat_long)
            site_dict['latitude']         = Angle(degrees=site_dict['latitude'])
            site_dict['longitude']        = Angle(degrees=site_dict['longitude'])
            site_dict['horizon']          = Angle(degrees=site_dict['horizon'])
            site_dict['ha_limit_neg']     = Angle(degrees=site_dict['ha_limit_neg'] * HOURS_TO_DEGREES)
            site_dict['ha_limit_pos']     = Angle(degrees=site_dict['ha_limit_pos'] * HOURS_TO_DEGREES)
            self.sites[site_dict['name']] = site_dict

        return


    def get(self, site_name):
        return self.sites.get(site_name)


    def __iter__(self):
        for site_dict in list(self.sites.values()):
            yield site_dict


def chunk_windows(window, chunksize):
    '''Given a user window, split that window into interval tuples (start, end)
       of size chunksize.'''
    chunk_start = window['start']
    chunk_end   = chunk_start + chunksize

    intervals = []
    while chunk_end <= window['end']:
        interval = (chunk_start, chunk_end)
        intervals.append(interval)

        chunk_start += chunksize
        chunk_end   += chunksize

    # Pick up any partial remaining interval
    if chunk_start < window['end']:
        interval = (chunk_start, window['end'])
        intervals.append(interval)

    return intervals


def find_moving_object_network_up_intervals(window, elements, site_filename, chunksize):
    '''Network wide wrapper for find_moving_object_up_intervals.'''

    sites = initialise_sites(site_filename)
    up_intervals_at = {}
    for site in sites:
        up_intervals, altitudes = find_moving_object_up_intervals(window, elements, site,
                                                                  chunksize)
        coalesced_intervals = coalesce_adjacent_intervals(up_intervals)
        up_intervals_at[site['name']] = coalesced_intervals

    return up_intervals_at


def hour_angle_within_limits(hour_angle, neg_limit, pos_limit):
   if ( hour_angle.in_degrees() > neg_limit.in_degrees()  and
        hour_angle.in_degrees() < pos_limit.in_degrees()):

       return True

   return False


def ephemeris_chunk_within_ha_limits(ha1, ha2, neg_limit, pos_limit):
    if ( hour_angle_within_limits(ha1, neg_limit, pos_limit) and
         hour_angle_within_limits(ha2, neg_limit, pos_limit) ):
           return True

    return False


def ephemeris_chunk_above_horizon(alt1, alt2, horizon):
    if (  alt1.in_degrees() > horizon.in_degrees()  and
          alt2.in_degrees() > horizon.in_degrees()  ):
        return True

    return False


def find_moving_object_up_intervals(window, elements, site, chunksize=timedelta(minutes=15)):
    '''Return only the (interval, altitude) pairs for which the moving object is
       above the horizon and within the hour angle limits at the provided site.
       If you just want rising and setting visible intevals, supply hour angle limits
       of -12 and 12 in your site dict.
    '''

    # Map the type of moving object to SLALIB numeric convention
    target_type = elements['type'].lower()
    if target_type == 'mpc_minor_planet':
        jform = 2
    elif target_type == 'mpc_comet':
        jform = 3
    else:
        raise MovingViolation("Unsupported target type: '%s'" % str(target_type))

    coords = calc_ephemerides(window, elements, site, chunksize, jform)

    intervals   = []
    altitudes   = []
    hour_angles = []
    for coord in coords:
        local_hour_angle = calc_local_hour_angle(coord['ra_app'],
                                                 site['longitude'],
                                                 coord['start'])
        altitude = calculate_altitude(site['latitude'].in_degrees(),
                                      coord['dec_app'].in_degrees(),
                                      local_hour_angle.in_degrees())

        interval = (coord['start'], coord['end'])
        altitudes.append(altitude)
        intervals.append(interval)
        hour_angles.append(local_hour_angle)

    up_intervals = []
    up_altitudes = []
    for i in range(len(intervals)-1):
        # Keep the interval only if both it and the next are up
        # (in other words, throw away partial intervals)
        # This works because we have an extra pseudo-interval with size 0
        if not ephemeris_chunk_above_horizon(altitudes[i], altitudes[i+1], site['horizon']):
            continue

        if not ephemeris_chunk_within_ha_limits(hour_angles[i], hour_angles[i+1],
                                                site['ha_limit_neg'],
                                                site['ha_limit_pos']):
            continue

        up_intervals.append(intervals[i])
        up_altitudes.append(altitudes[i])

    return up_intervals, up_altitudes


def calc_ephemerides(window, elements, site, chunksize=timedelta(minutes=15), JFORM=2):
    '''Return the apparent RA/Dec for the moving object specified by elements, for a given
       site, every chunksize minutes within the window.'''

    chunked_intervals = chunk_windows(window, chunksize)

    coords = []
    for start, end in chunked_intervals:
        ra_app, dec_app = elem_to_topocentric_apparent(start, elements, site, JFORM)
        ephem = {
                  'ra_app'  : ra_app,
                  'dec_app' : dec_app,
                  'start'   : start,
                  'end'     : end,
                }
        coords.append(ephem)

    start, end = chunked_intervals[-1]
    ra_app, dec_app = elem_to_topocentric_apparent(end, elements, site, JFORM)
    ephem = {
              'ra_app'  : ra_app,
              'dec_app' : dec_app,
              'start'   : end,
              'end'     : end,
            }
    coords.append(ephem)

    return coords



#!/usr/bin/env python

'''
utils.py - Low level rise_set utility methods

description

Author: Eric Saunders
        Jason Eastman
December 2013
'''

def coalesce_adjacent_intervals(intervals):
    '''Combine a set of datetime 2-tuples, coalescing adjacent intervals into
       larger intervals wherever possible.
    '''

    # Catch the special case where the target never rose
    if len(intervals) == 0:
        return intervals

    coalesced_intervals = [intervals[0]]
    for interval in intervals[1:]:

        # If the current interval end matches the next interval start...
        if coalesced_intervals[-1][1] == interval[0]:
            # ...the two intervals are contiguous - combine them
            coalesced_intervals[-1] = (coalesced_intervals[-1][0], interval[1])
        else:
            # ...the two intervals are not contiguous - store seperately
            coalesced_intervals.append(interval)

    return coalesced_intervals


def intersect_intervals(int1, int2):

    ''' Computes the intersections of two sets of datetime 2-tuples, each of which
    represents the overlap between the two tuples.
    '''
    intersect = []
    for start1, end1 in int1:
        for start2, end2 in int2:

            start0 = max(start1,start2)
            end0 = min(end1,end2)

            if start0 < end0:
                intersect.append((start0,end0))

    return intersect


def intersect_many_intervals(*args):

    ''' Generalizes intersect_intervals to an arbitrary number of lists
    e.g., when the object is up, sun is down, and hour angle within limits
    '''
    intersection = args[0]
    for interval in args:
        intersection = intersect_intervals(intersection,interval)

    return intersection


def is_moving_object(target):
    # If a type is not specified, default to sidereal objects
    if 'type' not in target:
        return False

    if target['type'].lower() in ('mpc_minor_planet', 'mpc_comet'):
        return True

    return False


class MovingViolation(Exception):
    '''Exception for moving object errors.'''
    pass



#!/usr/bin/env python

'''
visibility.py - Visibility interval calculations.

TODO: description

Author: Eric Saunders (esaunders@lcogt.net)
February 2011
'''

# Required for true (non-integer) division
from __future__ import division
from builtins import range
from builtins import object

# Standard libary imports
import datetime
import math
import copy

# Internal imports
from rise_set.astrometry     import (calc_sunrise_set, calc_planet_rise_set, calc_rise_set, RiseSetError,
                                     Star, gregorian_to_ut_mjd, ut_mjd_to_gmst, date_to_tdb, apparent_planet_pos,
                                     mean_to_apparent, angular_distance_between)
from rise_set.angle          import Angle
from rise_set.moving_objects import find_moving_object_up_intervals
from rise_set.utils          import (coalesce_adjacent_intervals, intersect_intervals,
                                     intersect_many_intervals, is_moving_object)

# Import logging modules
import logging

# Set the logger name
_log = logging.getLogger('rise_set.visibility')


# Set convenient constants
ONE_DAY  = datetime.timedelta(days=1)
MIDNIGHT = datetime.time()
# The average moon refraction from Astronomical Almanac A12: 34 minutes
MOON_REFRACTION = Angle(degrees=-0.5666667)


def set_airmass_limit(airmass, horizon):
    ''' Compare the provided maximum airmass limit with the horizon of a telescope, and
        return the effective horizon for rise/set purposes (whichever is higher elevation). If
        no airmass is provided, we default to the horizon.

        airmass = 1 / cos(zenith)
        horizon = 90 - zenith'''

    if not airmass:
        return horizon

    # We convert the horizon to airmass, not vice versa, to avoid small angle problems if a
    # very large airmass is provided
    zenith_distance = 90 - horizon
    horizon_airmass = 1 / math.cos(math.radians(zenith_distance))

    effective_horizon = horizon
    if airmass < horizon_airmass:
        effective_horizon = 90 - math.degrees(math.acos(1 / airmass))

    return effective_horizon


class Visibility(object):

    def __init__(self, site, start_date, end_date, horizon=0, twilight='sunrise',
                 ha_limit_neg=-4.9, ha_limit_pos=4.9):
        self.site         = site
        self.start_date   = start_date
        self.end_date     = end_date
        self.horizon      = Angle(degrees=horizon)
        self.twilight     = twilight

        if ha_limit_pos > 12.0 or ha_limit_pos < 0.0:
            msg = "Positive hour angle limit must fall between 0 and 12 hours"
            raise InvalidHourAngleLimit(msg)

        if ha_limit_neg < -12.0 or ha_limit_neg > 0.0:
            msg = "Negative hour angle limit must fall between -12 and 0 hours"
            raise InvalidHourAngleLimit(msg)

        self.ha_limit_neg = ha_limit_neg
        self.ha_limit_pos = ha_limit_pos

        self.dark_intervals = []
        self.moon_dark_intervals = []


    def get_dark_intervals(self):
        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of uninterrupted darkness. The set of tuples gives the complete dark
           intervals between the Visibility object's start and end date.
        '''

        # Don't compute this again if we've already done it
        if self.dark_intervals:
            return self.dark_intervals

        target = 'sun'

        self.dark_intervals = self.get_target_intervals(target, up=False)

        return self.dark_intervals


    def get_moon_dark_intervals(self):
        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of uninterrupted darkness from the moon. The set of tuples gives the complete
           moon dark intervals between the Visibility object's start and end date.
        '''

        # Don't compute this again if we've already done it
        if self.moon_dark_intervals:
            return self.moon_dark_intervals

        target = 'moon'

        self.moon_dark_intervals = self.get_target_intervals(target, up=False)

        return self.moon_dark_intervals


    def get_moon_distance_intervals(self, target, target_intervals, moon_distance=Angle(degrees=30), chunksize=datetime.timedelta(minutes=30)):
        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of time that the target is greater than moon_distance away from the moon.
        '''
        intervals = []

        for start, end in target_intervals:
            chunkstart = start
            chunkend = min(chunkstart + chunksize, end)
            while chunkstart != chunkend and chunkend <= end:
                # get the tdb date of the start time of the interval
                tdb = date_to_tdb(chunkstart)
                # get the apparent ra/dec for the target, and for the moon at this timestamp
                target_app_ra, target_app_dec = mean_to_apparent(target, tdb)
                moon_app_ra, moon_app_dec, diameter = apparent_planet_pos('moon', tdb, self.site)

                # call slalib to get the angular moon distance
                target_moon_dist = angular_distance_between(target_app_ra, target_app_dec, moon_app_ra, moon_app_dec)
                # if that moon distance is > the constraint, add this interval to final intervals
                if target_moon_dist.in_degrees() >= moon_distance.in_degrees():
                    intervals.append((chunkstart, chunkend))
                # increment the chunkstart/end up
                chunkstart = chunkend
                chunkend = min(chunkstart + chunksize, end)

        intervals = coalesce_adjacent_intervals(intervals)
        return intervals


    def get_target_intervals(self, target, up=True, airmass=None):
        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of uninterrupted time when the target was above the horizon (or below, if
           up=False). The set of tuples gives the complete target down intervals
           between the Visibility object's start and end date.
        '''
        effective_horizon = set_airmass_limit(airmass, self.horizon.in_degrees())

        # Handle Satellite objects by just returning the dark intervals because we don't support explicit ephemeris
        # calculations for these objects
        if 'type' in target and target['type'] == 'Satellite':
            intervals = self.get_dark_intervals()
        # Handle moving objects differently from stars
        elif is_moving_object(target):
            intervals = self.get_moving_object_target_intervals(target, effective_horizon)
        # The target has an RA/Dec
        else:
            intervals = self.get_ra_target_intervals(target, up, airmass, effective_horizon)

        return intervals

    def get_moving_object_target_intervals(self, target, effective_horizon):
        window = {
                   'start' : self.start_date,
                   'end'   : self.end_date,
                 }
        site = self.site.copy()
        site['horizon'] = Angle(degrees=effective_horizon)
        intervals, _ = find_moving_object_up_intervals(window, target, site)

        return intervals


    def get_ra_target_intervals(self, target, up, airmass, effective_horizon):
        star = Star(self.site['latitude'], target, effective_horizon)

        if up:
            day_interval_func = self.find_when_target_is_up
        else:
            day_interval_func = self.find_when_target_is_down

        # Find rise/set/transit for each day
        intervals = []
        current_date = self.start_date
        while current_date < self.end_date + ONE_DAY:
            one_day_intervals = day_interval_func(target, current_date, star, airmass)

            # Add today's intervals to the accumulating list of intervals
            intervals.extend(one_day_intervals)

            # Move on to tomorrow
            current_date += ONE_DAY

        # Collapse adjacent intervals into continuous larger intervals
        intervals = coalesce_adjacent_intervals(intervals)
        intervals = intersect_intervals(intervals, [(self.start_date,self.end_date)])

        return intervals


    def get_ha_intervals(self, target):

        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of uninterrupted time when the target was within the Hour Angle limits of the
           telescope. The set of tuples gives the complete in range intervals
           between the Visibility object's start and end date.
        '''
        SIDEREAL_SOLAR_DAY_RATIO = 1.002737909350
        SIDEREAL_SOLAR_DAY = datetime.timedelta(seconds=(ONE_DAY.total_seconds() / SIDEREAL_SOLAR_DAY_RATIO))

        earliest_date = self.start_date - SIDEREAL_SOLAR_DAY

        mjd    = gregorian_to_ut_mjd(earliest_date)
        gmst   = ut_mjd_to_gmst(mjd)

        # the rise time
        hour_rise = self.ha_limit_neg + target['ra'].in_hours() - \
            self.site['longitude'].in_hours() - gmst.in_hours()
        hour_rise /= SIDEREAL_SOLAR_DAY_RATIO

        # the set time
        hour_set  = self.ha_limit_pos + target['ra'].in_hours() - \
            self.site['longitude'].in_hours() - gmst.in_hours()
        hour_set /= SIDEREAL_SOLAR_DAY_RATIO

        current_rise = earliest_date + datetime.timedelta(hours=hour_rise)
        current_set = earliest_date + datetime.timedelta(hours=hour_set)

        # Find hour angle limits for each day
        intervals = []
        while current_set < (self.end_date + SIDEREAL_SOLAR_DAY):
            intervals.append((current_rise, current_set))
            current_rise += SIDEREAL_SOLAR_DAY
            current_set += SIDEREAL_SOLAR_DAY

        # do not exceed start/end dates
        intervals = coalesce_adjacent_intervals(intervals)
        intervals = intersect_intervals(intervals,[(self.start_date,self.end_date)])

        return intervals


    def get_observable_intervals(self, target, airmass=None, moon_distance=Angle(degrees=30)):
        '''Returns a set of datetime 2-tuples, each of which represents an interval
           of uninterrupted time when the target is observable (sun down, target up,
           target within the Hour Angle limits of the telescope, target at least moon_distance
           away from the moon).
        '''

        # get the intervals of each separately
        dark               = self.get_dark_intervals()
        above_horizon      = self.get_target_intervals(target, airmass=airmass)
        if moon_distance.in_degrees() <= 0.5 or 'ra' not in target:
            moon_avoidance = above_horizon
        else:
            moon_avoidance = self.get_moon_distance_intervals(target, above_horizon, moon_distance)
        if 'ra' in target:
            within_hour_angle = self.get_ha_intervals(target)
        else:
            # if the target type is such that there is no 'ra'/'dec' values, then we cannot calculate the ha intervals,
            # so we just use the target intervals instead (since they are all intersected together next). This is true
            # for moving objects and satellite objects.
            within_hour_angle = above_horizon

        # find the overlapping intervals between them
        intervals = intersect_many_intervals(dark, above_horizon, within_hour_angle, moon_avoidance)

        return intervals


    def find_when_target_is_down(self, target, dt, star=None, airmass=None):
        '''Returns a single datetime 2-tuple, representing an interval
           of uninterrupted time below the horizon at the specified site, for the
           requested date.

           Note: Even though this function currently ignores times, the dt object
           must be a datetime, *not* a date.
        '''

        # Ensure we only deal with dates, because our rise/set/transit tuple is
        # day specific.
        # TODO: Extend to arbitrary start/end times
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)

        # We will calculate down intervals as the inverse of the up intervals
        _log.debug("dt: %s", dt)
        up_intervals = self.find_when_target_is_up(target, dt, star, airmass)

        if not up_intervals:
            _log.warn("Got no up intervals!")
            _log.warn("dt was: %s", dt)
            _log.warn("target was: %s", target)

        down_intervals = []

        # If the first value has time 00:00:00, then the target starts up
        if up_intervals[0][0].time() == MIDNIGHT:
            pass

        # Otherwise the target starts down - so there's one extra interval at start
        else:
            down_start = dt
            down_end   = up_intervals[0][0]

            down_intervals.append((down_start, down_end))


        # Proceed through the intervals, extracting the gaps
        for i in range(len(up_intervals) - 1):
            down_start = up_intervals[i][1]
            down_end   = up_intervals[i+1][0]

            down_intervals.append((down_start, down_end))


        # If the target sets before the end of the day, grab that as an
        # extra down interval
        if up_intervals[-1][1].time() != MIDNIGHT:
            down_start = up_intervals[-1][1]
            down_end   = dt + ONE_DAY

            down_intervals.append((down_start, down_end))

        return down_intervals


    def find_when_target_is_up(self, target, dt, star=None, airmass=None):
        '''Returns a single datetime 2-tuple, representing an interval
           of uninterrupted time above the horizon at the specified
           site, for the requested date.

           Note: Even though this function currently ignores times, the dt
           object must be a datetime, *not* a date.

           TODO: Clean up this unpleasant interface - no star need be passed if
           the sun is the target, but one *must* be passed otherwise. This isn't
           obvious from the method signature.
        '''
        effective_horizon = set_airmass_limit(airmass, self.horizon.in_degrees())

        # Remove any time component of the provided datetime object
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)

        # Get the rise/set/transit times for the target, for this day
        intervals = []
        # TODO: Catch the RiseSetError for circumpolar stars
        # TODO: Return either a complete or empty interval, as appropriate
        # TODO: This requires either introspecting state in the error, or
        # TODO: calling an is_circumpolar method before this calculation

        if target == 'sun':
            transits, rises, sets = calc_sunrise_set(self.site, dt, self.twilight)
        elif target == 'moon':
            transits, rises, sets = calc_planet_rise_set(self.site, dt, MOON_REFRACTION, 'moon')
        else:
            # Test for circumpolarity
            if star.is_always_up(dt):
                # Return a full interval over the entire day
                intervals.append((dt, dt + ONE_DAY))
                return intervals

            # Catch target never rising
            try:
                effective_horizon_angle = Angle(degrees=effective_horizon)
                transits, rises, sets = calc_rise_set(target, self.site,
                                                      dt, effective_horizon_angle)
            except RiseSetError:
                return intervals


        _log.debug("latitude: %s",     self.site['latitude'].in_degrees())
        _log.debug("longitude: %s",    self.site['longitude'].in_degrees())
        _log.debug("twilight: %s",     self.twilight)
        _log.debug("dt: %s",           dt)
        _log.debug("rise: %s (%s)",    rises, dt + rises)
        _log.debug("transit: %s (%s)", transits, dt + transits)
        _log.debug("set: %s (%s)",     sets, dt + sets)

#        import ipdb; ipdb.set_trace()

        # Case 1: Overlapping start of day boundary
        # Target rose yesterday, and sets today. Rises again later today.
        #         |       x                                     |
        #         |    x     x                                  |
        #         | x           x                               | x
        #        x|                x                           x|
        #     x   |                   x                     x   |
        #   Rise  0hr  Transit       Set                  Rise 24hr
        if (rises > transits) and (sets > transits):

            # Store the first interval - start of day until target set
            absolute_set = sets + dt
            intervals.append((dt, absolute_set))

            # Store the second interval - target rise until end of day
            absolute_rise = rises + dt
            intervals.append((absolute_rise, dt + ONE_DAY))


        # Case 2: Rise, set and transit all fall within the day, in order
        # Target rises today, transits, and sets before the day ends
        #         |                      x                     |
        #         |                   x     x                  |
        #         |                x           x               |
        #         |             x                 x            |
        #         |          x                       x         |
        #         |       x                             x      |
        #         0hr   Rise           Transit          Set   24hr
        elif (rises < transits) and (sets > transits):
            # Only one interval - rise until target set
            absolute_rise = rises + dt
            absolute_set  = sets + dt
            intervals.append((absolute_rise, absolute_set))


        # Case 3: Overlapping end of day boundary
        # Target rose yesterday, and sets today. Rises again later today.
        #                 x    |                                        x    |
        #              x     x |                                     x     x |
        #           x          |x                                 x          |x
        #        x             |   x                           x             |   x
        #     x                |      x                     x                |
        #   Rise       Transit 0hr   Set                  Rise      Transit 24hr
        elif (rises < transits) and (sets < transits):
            # Same code as case 1!
            # Store the first interval - start of day until target set
            absolute_set = sets + dt
            intervals.append((dt, absolute_set))

            # Store the second interval - target rise until end of day
            absolute_rise = rises + dt
            intervals.append((absolute_rise, dt + ONE_DAY))


        return intervals


    def __repr__(self):
        repr_dict = copy.deepcopy(self.__dict__)
        del(repr_dict['dark_intervals'])

        sorted_str_dict = "{" + ", ".join("%s: %s" % (key, self.__dict__[key])
                             for key in sorted(self.__dict__)) + "}"
        return "Visibility (%s)" % sorted_str_dict

    def __key(self):
        return (self.site, self.start_date, self.end_date, self.horizon, self.twilight)

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __hash__(self):
        return hash(self.__key())



class InvalidHourAngleLimit(Exception):
    pass



#!/usr/bin/env python

'''This file controls the packaging and deployment of the rise_set module.

    Preferred Installation methods are found in the INSTALLATION file in this directory.

Eric Saunders
August 2012
'''

from setuptools import setup, find_packages

setup(
    name = 'rise_set',
    version = '0.4.2',
    description = 'Routines for accurate rise/set/transit calculations',
    author = 'Eric Saunders',
    author_email = 'esaunders@lcogt.net',
    packages = ['rise_set'],
    dependency_links = [
        "http://github.com/LCOGT/pyslalib/tarball/master#egg=pySLALIB",
    ],
    setup_requires = [
        "numpy"
    ],
    install_requires = [
        "pySLALIB",
        "numpy",
        "future",
    ],
    package_data = {
        '': ['*.conf'],
    },

    tests_require = [
        "mock",
        "pylint",
        "nose",
        "nosexcover"
    ],
    test_suite = 'nose.collector',
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
    ]

)



#!/usr/bin/env python

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import next
from builtins import zip
from builtins import str
from builtins import range
from past.utils import old_div
import ast
from datetime import datetime,timedelta
from visibility import Visibility
from sky_coordinates import RightAscension, Declination
from angle import Angle
import sys, argparse
from math import sqrt
from reqdb.utils.duration import calculate_duration
from reqdb.requests import Request, UserRequest
from reqdb.client   import SchedulerClient

# breaks up long intervals into discrete chunks that are schedulable
# by site with a minimum overlap of MINOVERLAP minutes and a maximum
# overlap of MAXOVERLAP minutes. Negative values of MINOVERLAP imply
# maximum gaps are acceptable. By default, it will schedule everything possible.

def distribute(interval, target, minoverlap=-1440, maxoverlap=1440, onepersite=True):

    sitenames = []
    intervals = []
    titles = []
    telescopes = file_to_dicts('/home/jeastman/lcogt/scheduler/trunk/test/telescopes.dat')
    for telescope in telescopes:
        if telescope['status'] == 'online':
            site =  {
                'name'     : telescope['name'].split(".")[2],
                'latitude' : Angle(degrees=telescope['latitude']),
                'longitude': Angle(degrees=telescope['longitude']),
                'horizon'  : telescope['horizon'],
                'ha_limit_neg' : telescope['ha_limit_neg'],
                'ha_limit_pos' : telescope['ha_limit_pos'],
                }            

            if not site['name'] in sitenames or not onepersite:
                sitenames.append(site['name'])
                print(site['name'])
                for obs_start,obs_end in interval:
                    observability = Visibility(
                        site=site,
                        start_date=obs_start,
                        end_date=obs_end,
                        horizon=site['horizon'],
                        twilight='nautical',
                        ha_limit_neg=site['ha_limit_neg'],
                        ha_limit_pos=site['ha_limit_pos'],
                        )
                    site_intervals = observability.get_observable_intervals(target, moon_distance=Angle(degrees=0))
                    for start,stop in site_intervals:
                        intervals.append((start,stop))
                        titles.append(target['name'] + ' (' + site['name'] + ')')

    return intervals, titles

# Read a file of code into a list (for telescopes.dat)
def file_to_dicts(filename):
    fh = open(filename, 'r')
    data = fh.read()
    return ast.literal_eval(data)

# read in the exoplanets.org csv file
def readexo():
    import csv
    f = open('exoplanets.csv','rb')
    reader = csv.reader(f)
    headers = next(reader)
    planets = {}
    for h in headers:
        planets[h] = []
    for row in reader:
        for h,v in zip(headers,row):
            planets[h].append(v)
    return planets


# get the number of exposures we can take during a given window for a single molecule
def get_nexp(window, molecule):

    diff = window['end'] - window['start']
    window_duration = old_div(diff.microseconds,1e6) + diff.seconds + diff.days*86400.0
    molecule['exposure_count'] = int(old_div(window_duration,molecule['exposure_time']))

    while True:
        molecule_duration = calculate_duration([molecule])
        if molecule_duration.seconds <= window_duration:
            break
        molecule['exposure_count'] -= 1


# Find the exoplanet parameters required to calculate the ephemeris from exoplanets.org
def get_ephem(name):

    # The name is specified, use values from exoplanets.org
    planets = readexo()
                    
    for i in range(len(planets['NAME'])):
        if planets['NAME'][i] == opt.name:
            # we should ensure these exist prior to using them
            per = float(planets['PER'][i])
            uper = float(planets['UPER'][i])
            tt = float(planets['TT'][i])
            utt = float(planets['UTT'][i])
            t14 = float(planets['T14'][i])
            ut14 = float(planets['UT14'][i])
            ra = RightAscension(degrees=float(planets['RA'][i])*15.0)
            dec = Declination(degrees=float(planets['DEC'][i]))
            vmag = float(planets['V'][i])

            target =  {
                'name'              : opt.name,
                'ra'                : ra,
                'dec'               : dec,
                'epoch'             : 2000,
                }
           
            return per, uper, tt, utt, t14, ut14, ra, dec, vmag
        
    # not found, return an error
    return -1

# calculates the transit intervals for all sites
def get_all_transit_intervals(target, ephemeris, start_date, end_date, onepersite=False, verbose=False, vverbose=False,exptime=0):
     
    intervals = []
    titles=[]
    sitenames = []
    telescopes = file_to_dicts('/home/jeastman/lcogt/scheduler/transits/rise_set-0.2.10/telescopes.dat')
    for telescope in telescopes:
        if telescope['status'] == 'online':
            site =  {
                'name'     : telescope['name'].split(".")[2],
                'latitude' : Angle(degrees=telescope['latitude']),
                'longitude': Angle(degrees=telescope['longitude']),
                'horizon'  : telescope['horizon'],
                'ha_limit_neg' : telescope['ha_limit_neg'],
                'ha_limit_pos' : telescope['ha_limit_pos'],
                }
            if not site['name'] in sitenames or not onepersite:
                sitenames.append(site['name'])

                site_intervals,site_titles = get_transit_intervals(target, ephemeris, site, start_date, end_date, verbose=verbose, vverbose=vverbose,exptime=exptime)

#                print site_intervals, site_titles, start_date, end_date, target
                for start,stop in site_intervals:
                    intervals.append((start,stop))
                for title in site_titles:
                    titles.append(title)

    return intervals,titles
    

def get_transit_intervals(target, ephemeris, site, start_date, end_date, verbose=False, vverbose=False,exptime=0):

    # period in days (float)
    # tt is a datetime object of the center of the transit (BJD_TDB)
    # t14 is the duration of transit, in days
    # target is a target object that must include ra, dec
    # start_date is a datetime object of the start of the window (UTC)
    # end_date is a datetime object of the end of the window (UTC)
    
    # calculate the epoch numbers that correspond to the observing window    
    startdiff = start_date-ephemeris['tt']
    startepoch = int(old_div((old_div(startdiff.microseconds,864e10) + old_div(startdiff.seconds,86400) + startdiff.days),ephemeris['period']))
    enddiff = end_date-ephemeris['tt']
    stopepoch = int(old_div((old_div(enddiff.microseconds,864e10) + old_div(enddiff.seconds,86400) + enddiff.days),ephemeris['period']))+1

    sitestr = site['name']
    visible_transits = []
    titles = []
    for epoch in range(startepoch, stopepoch+1):
        
        ## Uncertainty in the Transit Time of the Nth transit
        uttn = sqrt(ephemeris['utt']**2 + (epoch*ephemeris['uperiod'])**2 + ephemeris['ut14']**2)
        
        ## this defines the ideal window, including 3-sigma uncertainty
        contact1 = ephemeris['tt'] + timedelta(days=ephemeris['period']*epoch - 0.5*ephemeris['t14']) #- 3*uttn
        contact4 = ephemeris['tt'] + timedelta(days=ephemeris['period']*epoch + 0.5*ephemeris['t14']) #+ 3*uttn
        event_start = contact1 - timedelta(days=(ephemeris['maxpre' ])*ephemeris['t14'] + 3.0*uttn)
        event_stop  = contact4 + timedelta(days=(ephemeris['maxpost'])*ephemeris['t14'] + 3.0*uttn)
#        print (ephemeris['maxpre' ]),ephemeris['t14'], 3.0*uttn, event_stop-event_start, event_start,event_stop

        ## convert event_start, event_stop, contact1, contact4 from BJD_TDB to JD_UTC
        # ignore for now; this will introduce a systematic error of up to ~9 minutes (depending on target and time of year)
        
        # determine when the object is observable from this telescope
        observability = Visibility(
            site=site,
            start_date=max(event_start,start_date),
            end_date=min(event_stop,end_date),
            horizon=site['horizon'],
            twilight='nautical',
            ha_limit_neg=site['ha_limit_neg'],
            ha_limit_pos=site['ha_limit_pos'],
            )
        observable_intervals = observability.get_observable_intervals(target, moon_distance=Angle(degrees=0))

        for visible in observable_intervals:
    
            # the maximum we could observe it
            obs_start = max(event_start,visible[0])
            obs_end = min(event_stop,visible[1])
            
            # how much of the transit is visible?
            transitvis = min(contact4,visible[1])-max(contact1,visible[0])
            transitfrac = max(round(old_div((old_div(transitvis.microseconds,864e10) + old_div(transitvis.seconds,86400.0) + transitvis.days),ephemeris['t14']),3),0)

            # how much prior to ingress is visible?
            previs = min(contact1,visible[1]) - max(event_start,visible[0])
            prefrac = round(max(old_div((old_div(previs.microseconds,864e10) + old_div(previs.seconds,86400.0) + previs.days),ephemeris['t14']),0),3)

            # how much after egress is visible?
            postvis = min(event_stop,visible[1])-max(contact4,visible[0])
            postfrac = round(max(old_div((old_div(postvis.microseconds,864e10) + old_div(postvis.seconds,86400.0) + postvis.days),ephemeris['t14']),0),3)

            # how much out of transit total is visible?
            ootfrac = prefrac + postfrac

#            print verbose

            # does it meet the criteria?
            if transitfrac >= ephemeris['minevent'] and prefrac >= ephemeris['minpre'] and \
                    postfrac >= ephemeris['minpost'] and ootfrac >= ephemeris['minbaseline']:
                visible_transits.append((obs_start,obs_end))
                titles.append(target['name'] + ' ' + site['name'] + ' ' + str(epoch) + ' ' + str(transitfrac) + ' ' + str(ootfrac))
                if verbose or vverbose:
                    window = {
                        'start' : obs_start + timedelta(minutes=5),
                        'end' : obs_end - timedelta(minutes=5)
                        }    
                    molecule = {
                        # Required fields
                        'exposure_time'   : exptime,        # Exposure time, in secs
                        'exposure_count'  : 0,              # The number of consecutive exposures
                        'filter'          : '',           # The generic filter name
                        # Optional fields. Defaults are as below.
                        'type'            : 'EXPOSE',       # The type of the molecule
                        'ag_name'         : 'SCICAM',       # '' to let it resolve; 'SCICAM' for self-guiding
                        'ag_mode'         : 'Optional',
                        'instrument_name' : 'SCICAM',       # This resolves to the main science camera on the scheduled resource
                        'bin_x'           : 2,              # Your binning choice. Right now these need to be the same.
                        'bin_y'           : 2,
                        }
#                    get_nexp(window, molecule)
                    print(site['name'], epoch, obs_start, obs_end, contact1, contact4, transitfrac, ootfrac, uttn*1440.0)
            else:
                if vverbose:
                    print(site['name'], epoch, obs_start, obs_end, contact1, contact4, transitfrac, ootfrac, molecule['exposure_count'], uttn*1440.0, "*")

                
    return visible_transits,titles

def get_exptime(vmag):
    defocus = 0.0
    m0 = 12.0
    exptime = 25*10.0**(old_div((vmag-m0),2.5))
    exptime = min(round(exptime,0),300)
    while exptime < 30:
        exptime += 10.0
        defocus += 1.0
        
    return exptime, defocus

def schedule_intervals(intervals, target, band, vmag, propid, email, titles, verbose=False, exptime=None, defocus=None):
    constraints = {'max_airmass' : 2.0}

    location = {'telescope_class' : '1m0'}

    proposal = {
        'proposal_id'   : propid,
        'user_id'       : email,
        }


    if exptime == None:
        exptime,defocus = get_exptime(vmag)
    elif defocus==None:
        defocus = 0

    # if defocused, guide with the guider
    # otherwise, use the science camera
    if defocus > 0:
        guider = ''
    else:
        guider = 'SCICAM'

    molecule = {
        # Required fields
        'exposure_time'   : exptime,        # Exposure time, in secs
        'exposure_count'  : 0,              # The number of consecutive exposures
        'filter'          : band,           # The generic filter name
        # Optional fields. Defaults are as below.
        'type'            : 'EXPOSE',       # The type of the molecule
        'ag_name'         : 'SCICAM',       # '' to let it resolve; 'SCICAM' for self-guiding
        'ag_mode'         : 'Optional',
        'instrument_name' : guider,         # This resolves to the main science camera on the scheduled resource
        'bin_x'           : 2,              # Your binning choice. Right now these need to be the same.
        'bin_y'           : 2,
        'defocus'         : defocus         # Mechanism movement of M2, or how much focal plane has moved (mm)
        }
    
    # kludge because target changes definitions between Visibility and set_target
    target['ra'] = target['ra'].degrees
    target['dec'] = target['dec'].degrees
 
    i=0
    for obs_start,obs_end in intervals:

        # make the window a little shorter to calculate the number of exposures
        window = {
            'start' : obs_start + timedelta(minutes=2),
            'end' : obs_end - timedelta(minutes=2)
            }
        get_nexp(window, molecule)
        molecule['exposure_count'] -= 2

        req = Request()
        req.set_location(location)
        req.set_target(target)

        # this is the actual window
        window = {
            'start' : obs_start,
            'end' : obs_end
            }    
        req.add_window(window)

        req.add_molecule(molecule)
        req.set_constraints(constraints)
        
        ur = UserRequest(group_id=titles[i])
        ur.add_request(req)
        ur.operator = 'single'
        ur.set_proposal(proposal)

        if verbose:
            print(molecule['exposure_time'], molecule['exposure_count'], molecule['defocus'], vmag, band, titles[i])


        # You're done! Send the complete User Request to the DB for scheduling
        client        = SchedulerClient('http://scheduler-dev.lco.gtn/requestdb/')
        response_data = client.submit(ur, keep_record=True, debug=True)
        client.print_submit_response()
        i = i+1

    return

if __name__ == '__main__':

    # EXAMPLES:
    # see visible transits of WASP-12b in the next month
    # python transits2.py --name=WASP-12b

    # see visible transits of WASP-12b between 1/1/2013 and 1/5/2013
    # python transits2.py --name=WASP-12b --startwin=20130101 --stopwin=20130105

    # Schedule transits of WASP-12b between 1/1/2013 and 1/5/2013
    # python transits2.py --name=WASP-12b --startwin=20130101 --stopwin=20130105 --schedule --band=ip --email=email@lcogt.net --propid=XXXXX-###

    # Schedule transits of NEW_OBJECT between 1/1/2013 and 1/5/2013 using your own ephemeris and target info
    # python transits2.py --name=NEW_OBJECT --startwin=20130101 --stopwin=20130105 --period=1.09142245 --t14=0.122 --tt=2454508.976820 --ra=97.6366416613 --dec=29.672277778 --schedule --band=ip --email=email@lcogt.net --propid=XXXXX-###
    

    # define the arguments and defaults
    today = datetime.strftime(datetime.today()+timedelta(days=1), format='%Y%m%d')
    month = datetime.strftime(datetime.today()+timedelta(days=30), format='%Y%m%d')
    parser = argparse.ArgumentParser(description='Schedule transit observations on the LCOGT network')
    parser.add_argument('--name'         ,dest='name'         ,action='store'     ,type=str                ,help='Planet name from exoplanets.org from which to get ra, dec, period, uperiod, tt, utt, t14, ut14. Specifying those separately will override the exoplanets.org values. Spaces are ignored -- e.g., "WASP-12b" is equivalent to "WASP-12 b"')
    parser.add_argument('--period'       ,dest='per'          ,action='store'     ,type=float              ,help='Period of the planet (days).')
    parser.add_argument('--uperiod'      ,dest='uper'         ,action='store'     ,type=float              ,help='Uncertainty in the period (days)')
    parser.add_argument('--tt'           ,dest='tt'           ,action='store'     ,type=float              ,help='Transit time of the 0th epoch (BJD_TDB)')
    parser.add_argument('--utt'          ,dest='utt'          ,action='store'     ,type=float              ,help='Uncertainty in the transit time (days)')
    parser.add_argument('--t14'          ,dest='t14'          ,action='store'     ,type=float              ,help='Duration of the transit, from 1st to 4th contact (days)')
    parser.add_argument('--ut14'         ,dest='ut14'         ,action='store'     ,type=float              ,help='Uncertainty in the duration (days)')
    parser.add_argument('--ra'           ,dest='ra'           ,action='store'     ,type=float              ,help='Right ascension of the target (J2000 degrees)')
    parser.add_argument('--dec'          ,dest='dec'          ,action='store'     ,type=float              ,help='Declination of the target (J2000 degrees)')
    parser.add_argument('--mintransit'   ,dest='mintransit'   ,action='store'     ,type=float,default=1.0  ,help='Minimum fraction of the transit that must be visible to be scheduled')
    parser.add_argument('--minoot'       ,dest='minoot'       ,action='store'     ,type=float,default=0.5  ,help='Minimum fraction of the out of transit time that must be visible to be scheduled, in units of T14')
    parser.add_argument('--minpre'       ,dest='minpre'       ,action='store'     ,type=float,default=0.0  ,help='Minimum fraction of the out of transit before ingress that must be visible to be scheduled, in units of T14')
    parser.add_argument('--minpost'      ,dest='minpost'      ,action='store'     ,type=float,default=0.0  ,help='Minimum fraction of the out of transit after egress that must be visible to be scheduled, in units of T14')
    parser.add_argument('--maxpre'       ,dest='maxpre'       ,action='store'     ,type=float,default=0.5  ,help='Maximum fraction of the out of transit before ingress to attempt to schedule, in units of T14')
    parser.add_argument('--maxpost'      ,dest='maxpost'      ,action='store'     ,type=float,default=0.5  ,help='Maximum fraction of the out of transit after egress to attempt to schedule, in units of T14')
    parser.add_argument('--startwin'     ,dest='startwin'     ,action='store'     ,type=str  ,default=today,help='UT date of the beginning of the observing window (YYYYMMDD). Default is today.')
    parser.add_argument('--stopwin'      ,dest='stopwin'      ,action='store'     ,type=str  ,default=month,help='UT date of the end of the observing window (YYYYMMDD). Default is next month.')
    parser.add_argument('--schedule'     ,dest='schedule'     ,action='store_true'           ,default=False,help='Schedule these transits (requires BAND, PROPID, EMAIL, VMAG)')
    parser.add_argument('--band'         ,dest='band'         ,action='store'     ,type=str                ,help='The fitler to observe the transit')
    parser.add_argument('--propid'       ,dest='propid'       ,action='store'     ,type=str                ,help='The proposal id to associate with the scheduler')
    parser.add_argument('--email'        ,dest='email'        ,action='store'     ,type=str                ,help='The email associated with the proposal')
    parser.add_argument('--vmag'         ,dest='vmag'         ,action='store'     ,type=float              ,help='The absolute V band magnitude of the target')
    parser.add_argument('--verbose'      ,dest='verbose'      ,action='store_true'           ,default=False,help='Print extra information')
    parser.add_argument('--vverbose'     ,dest='vverbose'     ,action='store_true'           ,default=False,help='Print even more information (for debugging)')
    parser.add_argument('--onepersite'   ,dest='onepersite'   ,action='store_true'           ,default=False,help='Only schedule one transit at each site')
# these haven't been implemented yet, but are on my to-do list...
    parser.add_argument('--exptime'      ,dest='exptime'      ,action='store'     ,type=float              ,help='The desired exposure time (either this or VMAG required if SCHEDULE is set)')
    parser.add_argument('--defocus'      ,dest='defocus'      ,action='store'     ,type=float              ,help='The defocus, in mm. Only used if SCHEDULE is set. If not set, it is calculated based on VMAG')
#    parser.add_argument('--secondary'    ,dest='secondary'    ,action='store_true',           default=False,help='Calculate secondary eclipses, not primary transits')
#    parser.add_argument('--nschedule'    ,dest='nschedule'    ,action='store'     ,type=int                ,help='The maximum number of transits to schedule. Default is all.')
#    parser.add_argument('--schedulebest' ,dest='schedulebest' ,action='store_true'           ,default=False,help='Only schedule NSCHEDULE best (sorted by faction of transit visible, then by fraction of OOT visible). Default is soonest.')
#    parser.add_argument('--multiplesites',dest='multiplesites',action='store_true',           default=False,help='Allow observations at multiple sites to satisfy constraints')
#    parser.add_argument('--minoverlap'   ,dest='minoverlap'   ,action='store'     ,type=float,default=0.0  ,help='Minimum overlap between sites for each transit (hrs). Default is 0.')
#    parser.add_argument('--maxoverlap'   ,dest='maxoverlap'   ,action='store'     ,type=float,default=24.0 ,help='Maximum overlap between sites for each transit (hrs). Default is 24.')
#    parser.add_argument('--search'       ,dest='search'       ,action='store_true'           ,default=False,help='Schedule observations that cover the entire 3-sigma uncertainty window of a single transit (use to search for transits)')

    opt = parser.parse_args()

    per = None
    tt = None
    t14 = None
    ra = None
    dec = None
    
    # it's an error not to specify either the name or period, tt, t14, ra, and dec
    # we should check for that...

    # The name is specified, use values from exoplanets.org
    if opt.name != None:

        planets = readexo()
                    
        found = False
        for i in range(len(planets['NAME'])):
            if planets['NAME'][i].replace(' ','') == opt.name.replace(' ',''):

                # get the ephemeris values from exoplanets.org
                # if blank, set to None (to overwrite/check later)
                if planets['PER'][i] == '':
                    per = None
                else:
                    per = float(planets['PER'][i])
                if planets['UPER'][i] == '':
                    uper = 0.0
                else:
                    uper = float(planets['UPER'][i])
                if planets['TT'][i] == '':
                    tt = None
                else:                   
                    tt = float(planets['TT'][i])
                if planets['UTT'][i] == '':
                    utt = 0.0
                else:
                    utt = float(planets['UTT'][i])
                if planets['T14'][i] == '':
                    t14 = None
                else:
                    t14 = float(planets['T14'][i])
                if planets['UT14'][i] == '':
                    t14 = 0.0
                else:
                    ut14 = float(planets['UT14'][i])
                

                ra = RightAscension(degrees=float(planets['RA'][i])*15.0)
                dec = Declination(degrees=float(planets['DEC'][i]))
                vmag = float(planets['V'][i])
                print(vmag)
                found = True
                break
        if not found:
            print(opt.name + ' not found in exoplanets.org; using command line arguments')




    # Override exoplanets.org values with input values, if specified
    if opt.per != None:
        per = opt.per
    if opt.uper != None:
        uper = opt.uper
    elif uper == None:
        uper = 0.0

    if opt.vmag != None:
        vmag = opt.vmag
        
    if opt.tt != None:
        tt = opt.tt    
    if opt.utt != None:
        utt = opt.utt
    elif utt == None:
        utt = 0.0

    if opt.t14 != None:
        t14 = opt.t14
    if opt.ut14 != None:
        ut14 = opt.ut14
    else:
        ut14 = 0.0

    if opt.ra != None:
        ra = RightAscension(degrees=opt.ra)
    if opt.dec != None:
        dec = Declination(degrees=opt.dec)

#    if per==None or tt==None or t14==None or not ra<>None or not dec<>None:
#        print "ERROR: must specify either --name or per, tt, t14, ra, dec"
#        sys.exit(0)

    # Error checking
    if opt.schedule:
        if opt.band == None:
            print("Must specify --band if --schedule is set")
            sys.exit(0)
        if vmag == None:
            print("Must specify --vmag if --schedule is set")
            sys.exit(0)    
        if opt.propid == None:
            print("Must specify --propid if --schedule is set")
            sys.exit(0)
        if opt.email == None:
            print("Must specify --email if --schedule is set")
            sys.exit(0)

    print(vmag - 3)
            

    if per == None:
        print("Period not found; must specify --period")
        sys.exit(0)
    if tt == None:
        print("Transit time not found; must specify --tt")
        sys.exit(0)
    if t14 == None:
        print("Duration not found, must specify --t14")
        sys.exit(0)

#    if ra == None:
#        print "RA not found, must specify --ra"
#        sys.exit(0)
#    if len(dec) == 0:
#        print "Dec not found; must specify --dec"
#        sys.exit(0)

    if opt.vverbose:
        print(per, uper, tt, utt, t14, ut14, ra.degrees, dec.degrees)

    
    # define the target
    target = {
        'name'              : opt.name,
        'ra'                : ra,
        'dec'               : dec,
        'epoch'             : 2000,
        }

    # convert input dates to datetime objects
    start_date = datetime.strptime(opt.startwin,'%Y%m%d')
    end_date = datetime.strptime(opt.stopwin,'%Y%m%d')
    tt = datetime(1858,11,17,0,0,0) + timedelta(tt-2400000.5)

    ephemeris = {
        'tt'                : tt,
        'utt'               : utt,
        'period'            : per,
        'uperiod'           : uper,
        't14'               : t14,
        'ut14'              : ut14,
        'minevent'          : opt.mintransit,
        'minbaseline'       : opt.minoot,
        'minpre'            : opt.minpre,
        'minpost'           : opt.minpost,
        'maxpre'            : opt.maxpre,
        'maxpost'           : opt.maxpost,
        }

    print(ephemeris, target, ra.degrees)

    # get the transit intervals for each site
    intervals,titles = get_all_transit_intervals(target, ephemeris, start_date, end_date, onepersite=opt.onepersite, verbose=opt.verbose or not opt.schedule, vverbose=opt.vverbose, exptime=opt.exptime)
   
    # schedule the transits
    if opt.schedule:
        schedule_intervals(intervals, target, opt.band, vmag, opt.propid, opt.email, titles, verbose=opt.vverbose, exptime=opt.exptime, defocus=opt.defocus)
